/*
 *  filelog.c - Logging mechanism
 *
 *  Copyright (C) 1999 Hugo Haas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/syslog.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <stdlib.h>

#include "defines.h"
#include "log.h"

/* Definition of loginfo structures */

struct loginfo log;
struct loginfo *icmp_log = NULL, *tcp_log = NULL, *udp_log = NULL;

struct loginfo_collection *li_collection = NULL;

/* Mutex either for stdout or for a file */
pthread_mutex_t log_mutex = PTHREAD_MUTEX_INITIALIZER;

/*
 * stdout_loginfo
 *
 * When run with debug options, write everything to stdout (run_as_daemon is 0)
 */

void stdout_loginfo(struct loginfo *li) {

  li->log = (void *) filedesc_log;
  li->level_or_fd = 1;
  li->file = NULL;
  li->open = (void *) dummy_open;
  li->close = (void *) dummy_close;
}

/*
 * log_with_syslog
 *
 * Initializes a log_info structure using syslog
 */

void log_with_syslog(struct loginfo *li, const int level) {
  extern int run_as_daemon;

  if (!run_as_daemon) {
    stdout_loginfo(li);
    return;
  }
  li->log = (void *) syslog;
  li->level_or_fd = level;
  if (li->file != NULL)
    free(li->file);
  li->file = NULL;
  li->open = (void *) syslog_open;
  li->close = (void *) dummy_close;
}

/*
 * log_in_file
 *
 * Set a loginfo structure such as it will log things in a file
 */

void log_in_file(struct loginfo *li, char *filename) {
  extern int run_as_daemon;

  if (!run_as_daemon) {
    stdout_loginfo(li);
    return;
  }
  li->log = (void *) filedesc_log;
  li->level_or_fd = 0;
  li->open = (void *) file_open;
  li->close = (void *) file_close;
  li->file = strdup(filename);
}

/*
 * dummy_iolog_function
 *
 * Dummy function, #define'd as dummy_open and dummy_close
 */

void dummy_iolog_function(int *fd, const char *filename) {}

/*
 * syslog_open
 *
 * Open a connection to the system logging daemon. Register ippl as a daemon.
 */

void syslog_open(int *level, const char *filename) {

  openlog("ippl", 0, LOGFACILITY);
}

/*
 * log_entry
 *
 * Log an entry into a stream
 */

void log_entry(int fd, char *entry) {
  time_t current = time(NULL);
  char date[27];

  if (fd == 0)
    return;

  snprintf(date, 27, asctime(localtime(&current)));

  pthread_mutex_lock(&log_mutex);

  write(fd, date+4, strlen(date)-10);
  write(fd, " ", 1);
  write(fd, entry, (strlen(entry) < 1023) ? strlen(entry) : 1023 );
  write(fd, "\n", 1);

  pthread_mutex_unlock(&log_mutex);
}

/*
 * filedesc_log
 *
 * Log information into a file descriptor
 */

void filedesc_log(const int fd, char *format, ...) {
  char buffer[BUFFER_SIZE];
  va_list msg;

  va_start(msg, format);
  vsnprintf(buffer, BUFFER_SIZE, format, msg);
  va_end(msg);

  log_entry(fd, buffer);
}

/*
 * file_open
 *
 * Open a log file
 */

void file_open(int *fd, const char *filename) {

  *fd = open((const char *) filename, O_WRONLY | O_APPEND | O_CREAT, 0640);
  if (*fd == 0) {
    log.log(log.level_or_fd, "Can't open log file %s.", filename);
  }
}

/*
 * file_close
 *
 * Close a log file
 */

void file_close(int *fd, const char *filename) {
 
  fsync(*fd);
  close(*fd);
  *fd = 0;
}        

/*
 * open_log_structures
 *
 * Open all the loginfo structures
 */

void open_log_structures() {
  struct loginfo_collection *p;

  for (p=li_collection; p!= NULL; p=p->next) {
    p->info.open(&(p->info.level_or_fd), p->info.file);
  }
}

/*
 * close_log_structures
 *
 * Close all the loginfo structures
 */

void close_log_structures() {
  struct loginfo_collection *p;

  for (p=li_collection; p!= NULL; p=p->next) {
    p->info.close(&(p->info.level_or_fd), p->info.file);
  }
}

/*
 * destroy_log_structures
 *
 * Destroy all the loginfo structures
 */

void destroy_log_structures() {
  struct loginfo_collection *p;

  while ((p = li_collection) != NULL) {
    li_collection = li_collection->next;
    if (p->info.file != NULL)
      free(p->info.file);
    free(p);
  }
}

/*
 * add_log_structure
 *
 * Add a log structure matching a description if not existing.
 */

struct loginfo *add_log_structure(int level, char *filename) {
  struct loginfo_collection *p;

  if ((level == 0) && (filename == NULL))
    return NULL;

  if ((level == LOGLEVEL) && (filename == NULL))
    return &log;

  for (p=li_collection; p!= NULL; p=p->next) {
    if (level == 0) {
      /* Match name? */
      if (!strcmp(p->info.file, filename)) {
	return(&(p->info));
      }
    }
    else {
      /* Match loglevel? */
      if (p->info.level_or_fd == level) {
	return(&(p->info));
      }
    }
  }

  p = (struct loginfo_collection *) malloc(sizeof(struct loginfo_collection));
  if (p == NULL)
    fatal();
  if (level == 0)
    log_in_file(&(p->info), filename);
  else
    log_with_syslog(&(p->info), level);

  p->next = li_collection;
  li_collection = p;

  return(&(p->info));
}
